/*
 
VMeter Computer-side code. 7/16/2011
VMeter.net
 
Copyright (c) 2011 Curious Inventor, LLC

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 
 */

#import "AppController.h"
#import "CustomView.h"
#import "MAAttachedWindow.h"
#import "VariableStore.h"
#import "SoundC.h"
#include "hidapi.h"
#include <IOKit/hid/IOHIDManager.h>
#include <IOKit/hid/IOHIDKeys.h>




#define WCHAR_ENCODING NSUTF32LittleEndianStringEncoding
#define VENDOR_ID 0x03eb
#define DEVICE_ID 0x1234
#define MIN_PRESSURE_INPUT 40
#define MAX_PRESSURE_INPUT 100
#define MIN_PRESSURE_OUTPUT 0
#define MAX_PRESSURE_OUTPUT 127
#define PRESSURE_FILTER_COEF .1
#define RAW_SENSOR_OFFSET 20


enum LED_CTL_SOURCE {
    TOUCH_ONLY = 0,
    MIDI,
    SYS_VOLUME
    };


@implementation AppController

@synthesize debug_text_field;
@synthesize debug_text_field2;
@synthesize touchCtrlNumPopList;
@synthesize ctrlNumStringArray;

@synthesize ledSourcePopList;
@synthesize ledCtrlNumPopList;
@synthesize pressureCtrlNumPopList;
@synthesize brightnessCtrlNumPopList;


@synthesize touchToCtrlCheckBox;
@synthesize touchToVolCheckBox;
@synthesize touchToNoteCheckBox;
@synthesize pressureToCtrlCheckBox;
@synthesize ignoreTouchCheckBox;
@synthesize ledBrightnessFromCtrlCheckBox;
@synthesize rawSensorsToCtrlsCheckBox;
@synthesize midiChannelPopList;
@synthesize touchOnOffToCtrlsCheckBox;
@synthesize touchOnOffCtrlNumPopList;


@synthesize ledSourceIndex;
@synthesize touchRouterCtrlNum;
@synthesize ledCtrlNum;
@synthesize pressureCtrlNum;
@synthesize brightnessCtrlNum;
@synthesize midiChannelNum;
@synthesize touchOnOffCtrlNum;

hid_device *handle;
int usb_out_counter;
unsigned long my_timer = 0;
float filtered_touch = 0.0;
char note_on = 0;
int last_note_on_pitch;
float ave_pressure = 0.0;
int raw_sensors[8];
int raw_sensors_ref[8];



// main init function, setup the pop-up window menus, the VMeter USB communication timer,
// control over system volume, 
- (void)awakeFromNib
{
    
    // Setup the V in the statusbar and the pop-up window
    float height = [[NSStatusBar systemStatusBar] thickness];
	float width = 30;
    NSRect viewFrame = NSMakeRect(0, 0, width, height);
    statusItem = [[[NSStatusBar systemStatusBar] statusItemWithLength:width] retain];
    [statusItem setToolTip:@"VMeter: click to configure"];
      
    // setup the timer that calls the function that periodically talks to the VMeter
    NSTimer * myTimer;
    myTimer = [NSTimer scheduledTimerWithTimeInterval:.004 target:self selector:@selector(onTimer:) userInfo:NULL repeats: YES];
    [[NSRunLoop currentRunLoop] addTimer:myTimer  forMode:NSEventTrackingRunLoopMode]; 
	
    // system sound stuff (for using the VMeter to control the system volume, an interface's volume, or another audio interface's volume
    [[SoundC sharedInstance] enumerateSoundDevices];
	//[self buildInputMenu];
	[self buildOutputMenu];
    
    [debug_text_field setStringValue:@""];
    GLOBALVAR->brightness = 16; // VMeter brightness goes from 0 to 16, 0 being off.
    
    // popup window setup
    [statusItem setView:[[[CustomView alloc] initWithFrame:viewFrame controller:self] autorelease]];      
    
    // fill all the drop-down lists in the pop-up window (things like the MIDI Channel, MIDI ctrl numbers, etc.)
    ctrlNumStringArray = [[NSMutableArray alloc] initWithCapacity:128];
    int i;
    NSString *temp;
    for (i=0;i<128;i++) {
        temp = [NSString stringWithFormat:@"%d",i];
        [ctrlNumStringArray addObject:temp];
        [temp release];        
    }
    NSArray *temp_array = [[NSArray alloc] initWithArray:ctrlNumStringArray];	
    [touchCtrlNumPopList addItemsWithTitles:temp_array];
    [ledCtrlNumPopList addItemsWithTitles:temp_array];
    [pressureCtrlNumPopList addItemsWithTitles:temp_array];
    [brightnessCtrlNumPopList addItemsWithTitles:temp_array];
    [touchOnOffCtrlNumPopList addItemsWithTitles:temp_array];
    [temp_array release];
    [ctrlNumStringArray removeAllObjects];
    for (i=0;i<16;i++) {
        temp = [NSString stringWithFormat:@"%d",i];
        [ctrlNumStringArray addObject:temp];
        [temp release];        
    }
    NSArray *temp_array2 = [[NSArray alloc] initWithArray:ctrlNumStringArray];    
    [midiChannelPopList addItemsWithTitles:temp_array2];
    [temp_array2 release];    
    [ctrlNumStringArray removeAllObjects];
    [ctrlNumStringArray addObject:@"Touch Only"];
   // [ctrlNumStringArray addObject:@"System Volume"];
    [ctrlNumStringArray addObject:@"MIDI Controller"];    
    NSArray *temp_array3 = [[NSArray alloc] initWithArray:ctrlNumStringArray];
    [ledSourcePopList addItemsWithTitles:temp_array3];
    [ctrlNumStringArray release];
    [temp_array3 release];
    
	[NSApp setDelegate:self]; // standard app stuff?
}


// use this function to hide the popup window when another app is clicked on. Currently not used so the pop-up window stays
// on top of everything when you're using other programs. 
- (void)applicationDidResignActive:(NSNotification *)aNotification {
//	GLOBALVAR->clicked = NO;
//    [attachedWindow orderOut:self];
//	[[statusItem view] setNeedsDisplay:YES];
}


// volume slider moved, change lights on vmeter if volume control mode is enabled
- (IBAction) sliderMoved: (id) sender {
    float vol;
    vol = [slider floatValue]; // 0 to 100
	[[SoundC sharedInstance] setVolume:vol];
    if ([touchToVolCheckBox state]) {
        GLOBALVAR->last_led_setpoint = vol*2.55;
        [self send_column_to_USB:vol*2.55];
    }
}


// this is called before any usb read or write command to make sure we're connected, and
// connect if not.
// the handle variable gets set to NULL if VMeter is unplugged in the onTimer function that
// reads the packet from the VMeter
-(void) check_for_vmeter {
    wchar_t mystring[256];
    char buf[17];
    CFTypeRef ref;
    int result;
    if (!handle) {
        handle = hid_open(VENDOR_ID, DEVICE_ID, NULL);
        if (handle) {
            hid_set_nonblocking(handle, 1);
            hid_get_product_string(handle, mystring, 255);           
            NSString * temp_string = [[NSString alloc] initWithBytes:mystring length:(80) encoding:WCHAR_ENCODING];
            NSString * temp_string2 = [NSString stringWithFormat:@"Connected: Found %@", temp_string];             
            [debug_text_field setStringValue:temp_string2]; 
            
            while(hid_read(handle, (unsigned char *)buf, sizeof(buf)) != 16);            
       
            // store the calibration values and also the light level:
            // buf[2] has the current light setting, 0 thru 38.
            GLOBALVAR->last_led_setpoint = ((float)buf[2]) / 38.0 * 255.0;
            
            // we record the first sensor outputs as reference for sending out raw sensor values later on
            raw_sensors_ref[0] = buf[3];
            raw_sensors_ref[1] = buf[4];
            raw_sensors_ref[2] = buf[5];
            raw_sensors_ref[3] = buf[6];
            raw_sensors_ref[4] = buf[7];
            raw_sensors_ref[5] = buf[8];
            raw_sensors_ref[6] = buf[9];
            raw_sensors_ref[7] = buf[10];            
            
        }    
        else {  
            NSString *temp_string = [NSString stringWithFormat:@"Please connect VMeter"];
            [debug_text_field setStringValue:temp_string]; 
        }
    }
    else { 
        // added a callback inside hid.c to set a handle->alive variable to 0 when a hid device is unplugged. hid_read will
        // then return -5 from hid_read, which signals the calling function to set handle = NULL.
        // In summary, if handle is not NULL, we should be connected. TODO: There may be a race condition somewhere in here...
    }
}

// changes which side the light column is drawn from, top to bottom or bottom to top.
- (IBAction)upside_down_clicked:(id)sender {    
    int return_val;
    unsigned char buf[17];
    unsigned char but_state;    
    if ([upside_down state] == NSOffState) {
        but_state = 0;
    }
    else {
        but_state = 1;
    } 
    buf[0] = 0;
    buf[1] = 252; // 252 means buf[2] contains the upside-down state
    buf[2] = but_state;
    buf[3] = GLOBALVAR->brightness;
    [self check_for_vmeter]; // open connection if not already connected
    if (handle) {
        return_val = hid_write(handle, buf, 9);
    }  
    [self send_column_to_USB:GLOBALVAR->last_led_setpoint];      
}

// prep VMeter for firmware upgrade. Requires additional software and firmware. VMeter must be
// disconnected and reconnected to get out of this mode.
- (IBAction)upgrade_firmware:(id)sender {
    int return_val;
    unsigned char buf[17];    
    buf[0] = 0;
    buf[1] = 254; 
    buf[2] = 0;
    buf[3] = GLOBALVAR->brightness;
    [self check_for_vmeter];
    if (handle) {
        return_val = hid_write(handle, buf, 9);
    }   
}

// forces recalibration--shouldn't normally be necessary as it's constantly adjusting to the surrounding environment.
- (IBAction)recalibrate:(id)sender {
    int return_val;
    char buf[17];    
    buf[0] = 0;
    buf[1] = 253; 
    buf[2] = 0;
    buf[3] = GLOBALVAR->brightness;
    [self check_for_vmeter];
    if (handle) {
        return_val = hid_write(handle, buf, 9);
    }
    else 
        return;
    
    // pause for a sec:
    sleep(.2);
    
    // read back a packet to store reference values for the raw sensor outputs:
    while(hid_read(handle, (unsigned char *)buf, sizeof(buf)) != 16); 
    
    raw_sensors_ref[0] = buf[3];
    raw_sensors_ref[1] = buf[4];
    raw_sensors_ref[2] = buf[5];
    raw_sensors_ref[3] = buf[6];
    raw_sensors_ref[4] = buf[7];
    raw_sensors_ref[5] = buf[8];
    raw_sensors_ref[6] = buf[9];
    raw_sensors_ref[7] = buf[10];    
}

// this is the main communication with the VMeter.  This gets called every 4ms or so and reads a
// 16 byte packet from the VMeter.  The compelte protocol is detailed at the bottom of this file.
// Depending on what mode it's in, MIDI messages are triggered, the system volume is changed, etc.
-(void) onTimer: (NSTimer *)timer{
	int res, res2, i;
    int pressure_input;
    char buf[17];
    int val_to_send;
    unsigned char buf2[17];
    float temp;
    // if VMeter touched, set volume
    float v;
    int limited_pressure, scaled_pressure;    

    [self check_for_vmeter]; // connect if not already connected.
    if (handle) {       
       do { // since the mac could be buffering older messages, read them in until we get to 
           // the latest one, effectively throwing out the older ones.   
           [self check_for_vmeter];
           res2 = hid_read(handle, (unsigned char *)buf2, sizeof(buf2));
          // NSLog(@"hid_read: %d", res2);
           if (res2==16) { // valid, copy to buf
               for (i=0;i<16;i++) {
                   buf[i] = buf2[i];                   
               }  
               res = res2;
           } 
           else if (res2 == -5) {// connection is dead, VMeter has likely been unplugged
               handle = NULL;
               return;
           }
       }
       while(res2 == 16);
       if (res2 == -1) {
			handle = NULL;
            return;
		}     
       if (res == 16) { // packet is valid
//           packet[0]: Touch state, 0 for off, 1 for on
//           packet[1]: not used
//           packet[2]: current light position being displayed, 0 - 38
//           packet[3]: cap sensor 1 value, closest to VMeter label and USB port, can be negative
//           packet[4]: cap sensor 2 value, can be negative
//           packet[5]: cap sensor 3 value, can be negative
//           packet[6]: cap sensor 4 value, can be negative
//           packet[7]: cap sensor 5 value, can be negative           
//           packet[8]: cap sensor 6 value, can be negative
//           packet[9]: cap sensor 7 value, can be negative
//           packet[10]: cap sensor 8 value, can be negative
//           packet[11]: diagnostic, not specified
//           packet[12]: touch_pos (0-255)
//           packet[13]: pressure, or a sum of all the cap sensor differentials (current value - ref value). Can be negative?
//           packet[14]: not used
//           packet[15]: not used
           
           // send out raw sensor values if requested           
           if ([rawSensorsToCtrlsCheckBox state]) {
               for (i=0;i<8;i++) {
                   val_to_send = RAW_SENSOR_OFFSET - (buf[3+i]-raw_sensors_ref[i])*2.5;
                   if (val_to_send > 127)
                       val_to_send = 127;
                   if (val_to_send < 0)
                       val_to_send = 0;
                   [self sendCtrlMsg:val_to_send ctrl:50+i];                   
               }               
           }                
    
           if (buf[0] == 1) { // touch is active                
				temp = (float)((unsigned char)buf[12]) / 255.0; // normalize touch pos to 0-1
                GLOBALVAR->last_led_setpoint = temp*255.0;
                if ([touchToCtrlCheckBox state]) { // MIDI ctrl message for touch position 
                    [self sendCtrlMsg:(NSInteger)(temp*127.0) ctrl:touchRouterCtrlNum];               
                }
                if ([touchToVolCheckBox state]) { // change system volume
                    [slider setFloatValue:temp*100.0];
                    [[SoundC sharedInstance] setVolume:temp*100.0];
                }
                if ([pressureToCtrlCheckBox state]) {                    
                    pressure_input = (unsigned char)buf[13]; // the input from the vmeter will wrap if a hand is placed on all the sensors.  This will have to be protected against in the firmware.                  
                    // limiting:
                    limited_pressure = MIN(MAX(pressure_input,MIN_PRESSURE_INPUT),MAX_PRESSURE_INPUT);                 
                    // scalaing:
                    scaled_pressure = (float)(limited_pressure - MIN_PRESSURE_INPUT) / (float)(MAX_PRESSURE_INPUT - MIN_PRESSURE_INPUT) * (float)(MAX_PRESSURE_OUTPUT - MIN_PRESSURE_OUTPUT) + MIN_PRESSURE_OUTPUT;                  
                    // filtering:
                    ave_pressure = ave_pressure + (scaled_pressure - ave_pressure)*PRESSURE_FILTER_COEF;
                    //  NSLog(@"pres: %d %d %d %d", pressure_input,limited_pressure,scaled_pressure,(int)ave_pressure);
                    [self send_pressure_via_MIDI:(int)ave_pressure];
                }
                if ([touchOnOffToCtrlsCheckBox state]) { // MIDI ctrl message for touch on/off
                    if (!(GLOBALVAR->touch_is_active)) { // only send on a change of touch
                        [self send_touch_on_off_via_MIDI:127];
                    }
                }
               if ([touchToNoteCheckBox state]) {
                    if (note_on == 0) {
                        [self sendNoteOnMsg:(NSInteger)(temp*12.0 + 60) velocity:127];// one octave
                        note_on = 1;
                        last_note_on_pitch = (NSInteger)(temp*12.0 + 60);
                    }
               }
                GLOBALVAR->touch_is_active = YES;
                filtered_touch = filtered_touch + (temp - filtered_touch)*.1; // IIR filter on touch pos
			}// end if touched
			else { // no touch deteced
                // individual LED test. Writes a binary counter out to each of the 5 groups of 8 LEDs. See the function for 
                // an example of how to control every individual LED. 
                //[self send_individual_light_control];               
                
                if ([touchOnOffToCtrlsCheckBox state]) {
                    if ((GLOBALVAR->touch_is_active)) { // only send on a change of touch
                        [self send_touch_on_off_via_MIDI:0];
                    }
                }
                GLOBALVAR->touch_is_active = NO;
                if ([touchToNoteCheckBox state]) {
                    if (note_on == 1) {
                        [self sendNoteOffMsg:last_note_on_pitch];
                        note_on = 0;
                    }
                }
                if ([touchToVolCheckBox state]) { // if system volume control enabled, update lights with volume in case
                                                  // it changed outside of this program.
                    v = [[SoundC sharedInstance] getVolume] * 2.55;
                    if (v != GLOBALVAR->last_led_setpoint) {
                        GLOBALVAR->last_led_setpoint = v;
                        if ((v >= 0) && (v<=255)) {
                            [slider setFloatValue:(v/2.55)];  
                            [self send_column_to_USB:(v)];                    
                        }	
                    }
                }
			}		
		} // 16 sized packet (valid)	           
    }
    else { // no VMeter connected   
        GLOBALVAR->touch_is_active = NO;
        float v = [[SoundC sharedInstance] getVolume];
        if ((v >= 0) && (v<=100)) {
            [slider setFloatValue:v];                              
        }	       
    } 
}


// example function showing how to control every individual light. Lights are controlled with 5 bytes, the binary value
// of each byte control 8 individual LEDs. The 5th byte only controls 8 LEDs, 31-38.
char lights[5] = {0,0,0,0,0};
- (void)send_individual_light_control {
    int return_val;
    unsigned char buf[17];
    int j;    
    buf[0] = 0;
    buf[1] = 2; // this means buf[4-8] control individual lights, providing complete control over all 38 lights.
    buf[2] = 0; // ignored
    buf[3] = GLOBALVAR->brightness;
    buf[4] = lights[0];
    buf[5] = lights[1];
    buf[6] = lights[2];
    buf[7] = lights[3];
    buf[8] = lights[4];    
    NSLog(@"lights[0]: %d", lights[0]);
    for (j=0;j<5;j++) {
        lights[j]++;        
    }    
    [self check_for_vmeter];
    if (handle) {
        return_val = hid_write(handle, buf, 9);
    }    
}


// expects 0 to 255, draws a column from the bottom or top depending on the upside-down mode                            
- (void)send_column_to_USB:(int)v
{
    int return_val;
    unsigned char buf[17];
    buf[0] = 0;
    buf[1] = 1; // this means buf[2] has the volume, 0-255
    buf[2] = (unsigned char)(v);
    buf[3] = GLOBALVAR->brightness;

    [self check_for_vmeter];
    if (handle) {
        return_val = hid_write(handle, buf, 9);
    }    
}


// shows or hides the main settings windows when the V in the statusbar is clicked.
- (void)toggleAttachedWindowAtPoint:(NSPoint)pt
{
    // Attach/detach window.
    if (!attachedWindow) {
        attachedWindow = [[MAAttachedWindow alloc] initWithView:view 
                                                attachedToPoint:pt 
                                                       inWindow:nil 
                                                         onSide:MAPositionBottom 
                                                     atDistance:5.0];
		[attachedWindow setCollectionBehavior: NSWindowCollectionBehaviorCanJoinAllSpaces];
		[attachedWindow setLevel:NSFloatingWindowLevel];
		[attachedWindow setViewMargin:2];
		[attachedWindow setCornerRadius:8.0];
		[attachedWindow setBackgroundColor:[NSColor blackColor]];
		[attachedWindow setBackgroundColor:[NSColor colorWithCalibratedRed:0.0 green:0.0 blue:0.0 alpha:1.0]];
		[attachedWindow setBorderColor:[NSColor whiteColor]];
		//[self performSelector:@selector(onTimer:)];
			
		//[NSTimer scheduledTimerWithTimeInterval:0.01 target:self selector:@selector(onTimer:) userInfo:NULL repeats: YES];
	}	
    if ([attachedWindow isVisible] == NO) {
		//NSRect screenRect = [[NSScreen mainScreen] visibleFrame];
		//float w = MIN(screenRect.size.width - 20, 1024);
		//float h = MIN(screenRect.size.height - 40, 768);
		//[view setFrameSize:NSMakeSize(w, h)];
		[attachedWindow _redisplay];
        [attachedWindow makeKeyAndOrderFront:self];
		[NSApp activateIgnoringOtherApps:YES];
    } else {
        [attachedWindow orderOut:self];
    }    
}

- (void)drawStatusRect:(NSRect)rect High:(BOOL)hi {
	[statusItem drawStatusBarBackgroundInRect:rect withHighlight:hi];    
}


// additional code for detected and enumerating different audio devices.  For instance, you could change the volume of a 
// USB audio device.
/*
	//SELECTED DEVICES
	[[SoundC sharedInstance] getSelectedDevices];
	//SET INPUT MENU SELECTED ITEM
	NSString *deviceInputName = [NSString stringWithCString:[SoundC sharedInstance]->selectedInputDevice encoding:NSASCIIStringEncoding];
	for (int i = 0; i < [mInputDevices numberOfItems]; i++) {
		if ([[[mInputDevices itemAtIndex:i] title] isEqualToString:deviceInputName]) {
			[mInputDevices selectItemAtIndex:i];
			break;
		}
	}
	//SET OUTPUT MENU SELECTED ITEM
	NSString *deviceOutputName = [NSString stringWithCString:[SoundC sharedInstance]->selectedOutputDevice encoding:NSASCIIStringEncoding];
	for (int i = 0; i < [mOutputDevices numberOfItems]; i++) {
		if ([[[mOutputDevices itemAtIndex:i] title] isEqualToString:deviceOutputName]) {
			[mOutputDevices selectItemAtIndex:i];
			break;
		}
	}
}
- (IBAction)inputDeviceSelected:(id)sender {
	[[SoundC sharedInstance] setDeviceNamed:[mInputDevices titleOfSelectedItem] Type:kAudioTypeInput];
}

- (IBAction)outputDeviceSelected:(id)sender {
	[[SoundC sharedInstance] setDeviceNamed:[mOutputDevices titleOfSelectedItem] Type:kAudioTypeOutput];	
}

- (void) buildInputMenu {
	[mInputDevices removeAllItems];
	if ([SoundC sharedInstance]->lastDevInput < 0) return;
	for (int i = 0; i <= [SoundC sharedInstance]->lastDevInput; i++) {
		NSString *mName = [NSString stringWithCString:&[SoundC sharedInstance]->devInput[i].mName[0] encoding:NSASCIIStringEncoding];
		[mInputDevices addItemWithTitle:mName];
	}
}
 
 - (void)applicationDidFinishLaunching:(NSNotification *)aNotification
 {
     // Insert code here to initialize your application
     [NSTimer scheduledTimerWithTimeInterval:0.1 target:self selector:@selector(onTimer:) userInfo:NULL repeats: YES];
     [[SoundC sharedInstance] enumerateSoundDevices];
     //[self buildInputMenu];
     [self buildOutputMenu];
 }
 */
// detects all the audio output devices.
- (void) buildOutputMenu {
	[mOutputDevices removeAllItems];
	if ([SoundC sharedInstance]->lastDevOutput < 0) return;
	for (int i = 0; i <= [SoundC sharedInstance]->lastDevOutput; i++) {
		NSString *mName = [NSString stringWithCString:&[SoundC sharedInstance]->devOutput[i].mName[0] encoding:NSASCIIStringEncoding];
		[mOutputDevices addItemWithTitle:mName];
	}
}

// expects 0 to 16, 0 is off.
- (void)send_brightness_to_USB:(int)b {
    unsigned char buf[17];
    int return_val;
   // NSLog(@"set brightness: %d and vol: %d", b, GLOBALVAR->last_led_setpoint);     
    GLOBALVAR->brightness = b;
    [slider_brightness setFloatValue:b];    
    buf[0] = 0;
    buf[1] = 0; // this means buf[2] is ignored
    buf[2] = 0; //GLOBALVAR->last_led_setpoint;
    buf[3] = GLOBALVAR->brightness;    
    [self check_for_vmeter];
    if (handle) {
        return_val = hid_write(handle, buf, 9);
    }
}

- (IBAction)slider_brightness_moved:(id)sender {
    unsigned char b;
    b = (unsigned char)[slider_brightness floatValue];   
    GLOBALVAR->brightness = b;    
    [self send_brightness_to_USB: b]; 
}


- (IBAction)open_help:(id)sender {    
    NSWorkspace * ws = [NSWorkspace sharedWorkspace];
    NSURL *url = [NSURL URLWithString: @"http://vmeter.net/instructions"];
    [ws openURL:url];
}




// ------------------------------------------- //
// MIDI Stuff below:
- (id) init	{
	NSLog(@"AppController:init:");
	self = [super init];
	msgArray = [[NSMutableArray arrayWithCapacity:0] retain];
    midiManager = [[VVMIDIManager alloc] init];
    [midiManager setDelegate:self];
    [midiManager virtualDest].name = @"VMeter";
  //  NSLog(@"virtualDest: %@", [midiManager virtualDest].name);
	return self;
}
// when MIDI devices are added or taken away
- (void) setupChanged	{
//	NSLog(@"AppController:setupChanged:");
	[sourcesTableView reloadData];
	[receiversTableView reloadData];
}


// this gets called when MIDI comes in. It generates a list of messages for the MIDI input display, and acts on the MIDI messages
// according to the settings to change the VMeter LEDs.
- (void) receivedMIDI:(NSArray *)a fromNode:(VVMIDINode *)n	{
//	NSLog(@"AppController:receivedMIDI:");
//	NSLog(@"virtualDest: %@", [midiManager virtualDest].name);
    NSEnumerator		*it = [a objectEnumerator];
	VVMIDIMessage		*msgPtr;
	NSMutableString		*mutString = [NSMutableString stringWithCapacity:0];
    int ctrl_num, value, channel;    
	@synchronized(msgArray)	{
		//	add the received messages to the array
		while ((msgPtr = [it nextObject])) {
            channel = [msgPtr channel];
            ctrl_num = [msgPtr data1];
            value = [msgPtr data2];
            if (GLOBALVAR->touch_is_active == NO || GLOBALVAR->ignore_touch == YES) { // change lights if a finger isn't on it or ignore touch is enabled.
                if (channel == midiChannelNum) { // ignore other channels
                    if (ctrl_num == ledCtrlNum) {                    
                        if (ledSourceIndex == MIDI) { // if MIDI control over LEDs enabled
                            [self send_column_to_USB:value*2];
                        }                        
                    }
                    if (ctrl_num == brightnessCtrlNum) {           
                       // test = [ledBrightnessFromCtrlCheckBox state];
                        if ([ledBrightnessFromCtrlCheckBox state]) {
                            [self send_brightness_to_USB:(value+1)/8];
                        }                       
                    }
                }
            }            
			[msgArray addObject:[msgPtr lengthyDescription]];            
        }
		//	make sure the array only has 10 objects
		while ([msgArray count] > 10)
			[msgArray removeObjectAtIndex:0];
		//	assemble a string from the msg array
		it = [msgArray reverseObjectEnumerator];
		while ((msgPtr = [it nextObject])) {
			[mutString appendFormat:@"%@\n",msgPtr];            
        }
		//	push the string to the view
		[receivedView performSelectorOnMainThread:@selector(setString:) withObject:[[mutString copy] autorelease] waitUntilDone:NO];
	}
}

- (void) send_touch_on_off_via_MIDI: (NSInteger)val {
    VVMIDIMessage		*msg = nil;    
    msg = [VVMIDIMessage createFromVals:
                 VVMIDIControlChangeVal:
           [channelField intValue]:
                        touchOnOffCtrlNum:
           val];
	if (msg != nil)
		[midiManager sendMsg:msg];      
}
- (void) send_pressure_via_MIDI: (NSInteger)pres {
   	VVMIDIMessage		*msg = nil;    
    msg = [VVMIDIMessage createFromVals:
                 VVMIDIControlChangeVal:
           [channelField intValue]:
                     pressureCtrlNum:
           pres];
	if (msg != nil)
		[midiManager sendMsg:msg];  
}
- (void) sendCtrlMsg:(NSInteger)val ctrl:(NSInteger)cNum {
  //  NSLog(@"Ctrl: %d val: %d",cNum,val);
    VVMIDIMessage		*msg = nil;    
    msg = [VVMIDIMessage createFromVals:
                 VVMIDIControlChangeVal:
           midiChannelNum:
                     cNum:
           val];
	if (msg != nil)
		[midiManager sendMsg:msg];    
}
- (void) sendNoteOffMsg: (NSInteger)pitch {
    VVMIDIMessage		*msg = nil;  
     NSLog(@"noteoff pitch: %i", pitch);
    msg = [VVMIDIMessage createFromVals:
                    VVMIDINoteOffVal:
       [channelField intValue]:
                              pitch:
       0];
    if (msg != nil)
        [midiManager sendMsg:msg];   
}
- (void) sendNoteOnMsg: (NSInteger)pitch velocity:(NSInteger)vel {
    VVMIDIMessage		*msg = nil;    
    NSLog(@"pitch: %i", pitch);
    msg = [VVMIDIMessage createFromVals:
                 VVMIDINoteOnVal:
           [channelField intValue]:
                     pitch:
           vel];
	if (msg != nil)
		[midiManager sendMsg:msg];       
}

// MIDI device listings
- (NSInteger) numberOfRowsInTableView:(NSTableView *)tv	{
	if (tv == sourcesTableView)
		return [[midiManager sourceArray] count];
	else if (tv == receiversTableView)
		return [[midiManager destArray] count];
	
	return 0;
}
- (id) tableView:(NSTableView *)tv objectValueForTableColumn:(NSTableColumn *)tc row:(NSInteger)row	{
	if (tv == sourcesTableView)	{
		//NSLog(@"\t\tsources");
		id			midiSource = [[midiManager sourceArray] lockObjectAtIndex:row];
		
		if (midiSource == nil)
			return nil;
		
		if (tc == sourcesNameColumn)	{
			return [midiSource name];
		}
		else if (tc == sourcesEnableColumn)	{
			if ([midiSource enabled])
				return [NSNumber numberWithInt:NSOnState];
			else
				return [NSNumber numberWithInt:NSOffState];
		}
	}
	else if (tv == receiversTableView)	{
		//NSLog(@"\t\treceivers");
		id			midiSource = [[midiManager destArray] lockObjectAtIndex:row];
		
		if (midiSource == nil)
			return nil;
		
		if (tc == receiversNameColumn)	{
			return [midiSource name];
		}
		else if (tc == receiversEnableColumn)	{
			if ([midiSource enabled])
				return [NSNumber numberWithInt:NSOnState];
			else
				return [NSNumber numberWithInt:NSOffState];
		}
	}	
	return nil;
}
- (void) tableView:(NSTableView *)tv setObjectValue:(id)v forTableColumn:(NSTableColumn *)tc row:(NSInteger)row	{
	//NSLog(@"AppController:tableView:setObjectValue:forTableColun:row:");
	if (tv == sourcesTableView)	{
		//NSLog(@"\t\tsources");
		id			midiSource = [[midiManager sourceArray] lockObjectAtIndex:row];
		
		if (midiSource == nil)
			return;
		
		if (tc == sourcesEnableColumn)	{
			if ([v intValue] == NSOnState)
				[midiSource setEnabled:YES];
			else
				[midiSource setEnabled:NO];
		}
	}
	else if (tv == receiversTableView)	{
		//NSLog(@"\t\treceivers");
		id			midiSource = [[midiManager destArray] lockObjectAtIndex:row];
		
		if (midiSource == nil)
			return;
		
		if (tc == receiversEnableColumn)	{
			if ([v intValue] == NSOnState)
				[midiSource setEnabled:YES];
			else
				[midiSource setEnabled:NO];
		}
	}
}


// CHECK BOXES:
- (IBAction) touchRouteCtrlNumChange: (id)sender {
    touchRouterCtrlNum = [touchCtrlNumPopList indexOfSelectedItem];
    //NSLog(@"num :%i",[touchCtrlNumPopList indexOfSelectedItem]);    
}
- (IBAction)ledSourceChange: (id) sender {
    ledSourceIndex = [ledSourcePopList indexOfSelectedItem];
    NSLog(@"ledsourcechange :%i",ledSourceIndex);    
}
- (IBAction)ledCtrlNumChange: (id) sender {
    ledCtrlNum = [ledCtrlNumPopList indexOfSelectedItem];
    NSLog(@"ledCtrl: %i",ledCtrlNum);
}
- (IBAction)presCtrlNumChange: (id) sender {
    pressureCtrlNum = [pressureCtrlNumPopList indexOfSelectedItem];
    NSLog(@"pressureCtrlNum: %i",pressureCtrlNum);    
}
- (IBAction)brightnessCtrlNumChange: (id) sender {
    brightnessCtrlNum = [brightnessCtrlNumPopList indexOfSelectedItem];
    NSLog(@"brightnessCtrlNum: %i",brightnessCtrlNum);    
}
- (IBAction)midiChanNumChange: (id) sender {
    midiChannelNum = [midiChannelPopList indexOfSelectedItem];
    NSLog(@"midiChannelNum: %i",midiChannelNum);    
}
- (IBAction)touchOnOffCtrlNumChange: (id) sender {
    touchOnOffCtrlNum = [touchOnOffCtrlNumPopList indexOfSelectedItem];
    NSLog(@"touchOnOffCtrlNum: %i",touchOnOffCtrlNum);    
}


// not shown in current panel, but example of adding another slider in the interface
- (IBAction) ctrlValSliderUsed:(id)sender	{
    //	NSLog(@"AppController:ctrlValSliderUsed:");
	VVMIDIMessage		*msg = nil;    
    msg = [VVMIDIMessage createFromVals:
                 VVMIDIControlChangeVal:
           [channelField intValue]:
                     touchRouterCtrlNum:
           floor(127*[sender floatValue])];
	
    //	tell the midi manager to send it
	if (msg != nil)
		[midiManager sendMsg:msg];
}
// this sets the VMeter so touches are ignored. If ignore mode isn't turned on, touching the VMeter will override whatever light
// control signals are being sent via USB.
- (IBAction)ignoreTouchChange: (id) sender {
    unsigned char buf[9];
    int return_val;
    NSLog(@"IgnoreTouch Changed: %d", [ignoreTouchCheckBox state]);        
    buf[0] = 0;
    buf[1] = 251; // this means buf[2] controls whether the lights ignore touch or not
    buf[2] = !([ignoreTouchCheckBox state]); //GLOBALVAR->last_led_setpoint;
    buf[3] = GLOBALVAR->brightness;    
    [self check_for_vmeter];
    if (handle) {
        return_val = hid_write(handle, buf, 9);
    }  
    GLOBALVAR->ignore_touch = [ignoreTouchCheckBox state];
}

- (void)dealloc
{
    [[NSStatusBar systemStatusBar] removeStatusItem:statusItem];
    [midiManager release];
    [touchCtrlNumPopList release];
    [ledSourcePopList release];
    [super dealloc];
}



@end
    /*       
--------------------------------------------------     
VMeter USB Protocol
--------------------------------------------------
     
                      
The VMeter continously sends packets of 16 bytes and receives packets of 8 bytes about every 4ms (subject to change).

Some of the bytes sent from the VMeter are diagnostic and may change.

8 BYTE PACKET SENT TO VMETER:
     
packet[0]: always required, not actually sent to VMeter but must be 0.     
packet[1]: command (see below)
packet[2]: command parameter
packet[3]: brightness: 0-16, 0 is off, 16 is max brightness
packet[4]: binary value of lights 0-7 (when packet[1] == 2)
packet[5]: binary value of lights 8-15 (when packet[1] == 2)
packet[6]: binrary value of lights 16-23 (when packet[1] == 2)
packet[7]: binrary value of lights 24-31 (when packet[1] == 2)
packet[8]: binrary value of lights 32-38 (38 lights in total) (when packet[1] == 2)
     
commands:
     1: When packet[1]==1, packet[2] contains the column height, 0 thru 38. If upside-down mode is on, it draws from the top
     2: When packet[1]==2, packet[4-8] contain the binary value for each group of 8 lights. For instance, sending a 6 in packet[4]
        would turn on the 2nd and 3rd lights since 6 == 0b00000110
     251: When packet[1]==251, packet[2] sets the ignoreTouch mode. If 1, touches don't change the lights. If 0, touches override
        whatever is currently controlling the lights (MIDI for instance).
     252: packet[2] turns on / off upside-down mode.
     253: recalibrate.  Sometimes this might be necessary if you move the VMeter.
     254: Prepare to update firmware. Puts the VMEter in a special state that will accept new firmware. Requires additional software
            and firmware
     
     
    
16 BYTE PACKET RECEIVED FROM VMETER:

packet[0]: Touch state, 0 for off, 1 for on
packet[1]: not used
packet[2]: current light position being displayed, 0 - 38
packet[3]: cap sensor 1 value, closest to VMeter label and USB port, can be negative
packet[4]: cap sensor 2 value, can be negative
packet[5]: cap sensor 3 value, can be negative
packet[6]: cap sensor 4 value, can be negative
packet[7]: cap sensor 5 value, can be negative
     
packet[8]: cap sensor 6 value, can be negative
packet[9]: cap sensor 7 value, can be negative
packet[10]: cap sensor 8 value, can be negative
packet[11]: diagnostic, not specified
packet[12]: touch_pos (0-255)
packet[13]: pressure, or a sum of all the cap sensor differentials (current value - ref value). Can be negative?
packet[14]: not used
packet[15]: not used
     

     
     */
